package imcacat.adxvctl;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.apache.log4j.Logger;

import imcacat.epicsca4j.model.EpicsPv;
import imcacat.epicsca4j.util.ObjectUtilities;

import com.jgoodies.binding.beans.Model;
import com.jgoodies.binding.value.ValueModel;

/**
 * I am the model for this application.
 */
public class Adxvctl extends Model {
  private static final long serialVersionUID = 1;
  private static final Logger LOG = Logger.getLogger(Adxvctl.class);

  protected static final File NONEXISTENT_FILE = new File("/nonexistent");

  public static final String PROPNAME_COLLECTED_IMAGE = "collectedImage";
  public static final String PROPNAME_ADXV_IMAGE = "adxvImage";
  public static final String PROPNAME_UPDATE_EVERY_ENABLED =
    "updateEveryEnabled";
  public static final String PROPNAME_UPDATE_EVERY_INTERVAL =
    "updateEveryInterval";
  public static final String PROPNAME_USER_MESSAGE = "userMessage";

  public static final String USER_MESSAGE_PREFIX_ERROR = "Error: ";
  public static final String USER_MESSAGE_PREFIX_WARN = "Warning: ";

  protected List<EpicsPv> collectedImagePathPvs = null;
  protected File collectedImage = null;
  protected File adxvImage = null;
  protected long adxvImageCount = 1;
  protected boolean updateEveryEnabled = true;
  protected int updateEveryInterval = 1;
  protected int updateEveryCount = updateEveryInterval;
  protected String userMessage = "";
  protected WrittenToDiskMonitor writtenToDiskMonitor =
    new WrittenToDiskMonitor();
  protected Thread writtenToDiskMonitorThread = new Thread(
    writtenToDiskMonitor);

  public Adxvctl(String collectedImagePathPvName) {
    this(Arrays.asList(collectedImagePathPvName));
  }

  public Adxvctl(List<String> collectedImagePathPvNames) {
    collectedImagePathPvs = new ArrayList<EpicsPv>(
        collectedImagePathPvNames.size());
    for (String each : collectedImagePathPvNames) {
      EpicsPv eachPv = new EpicsPv(each, true, true,
        new ByteArrayToStringEpicsCaValueConverter());
      eachPv.addPropertyChangeListener(ValueModel.PROPERTY_VALUE,
        new PropertyChangeListener() {
          public void propertyChange(PropertyChangeEvent event) {
            collectedImagePathChanged((String)event.getNewValue());
          }
        });
      collectedImagePathPvs.add(eachPv);
    }

    writtenToDiskMonitor.addWrittenToDiskListener(
      new WrittenToDiskListener() {
        public void writtenToDisk(File image) {
          imageWrittenToDisk(image);
        }});

    writtenToDiskMonitorThread.setDaemon(true);
    writtenToDiskMonitorThread.start();
  }

  public void release() {
    writtenToDiskMonitorThread.interrupt();
  }

  public File getCollectedImage() {
    return collectedImage;
  }

  protected void setCollectedImage(File newValue) {
    File oldValue = collectedImage;
    collectedImage = newValue;
    if (ObjectUtilities.areEqual(oldValue, newValue)) return;
    File oldParent = (oldValue == null) ? null : oldValue.getParentFile();
    File newParent = (newValue == null) ? null : newValue.getParentFile();
    if (!ObjectUtilities.areEqual(oldParent, newParent)) {
      updateEveryCount = updateEveryInterval;
    }
    firePropertyChange(Adxvctl.PROPNAME_COLLECTED_IMAGE, oldValue, newValue);
    if (!updateEveryEnabled) return;
    if (updateEveryCount >= updateEveryInterval) {
      boolean isFirstImage = oldValue == null;
      enqueueAdxvImageUpdate(newValue,
        isFirstImage ? 0 : System.currentTimeMillis());
      updateEveryCount = 1;
    } else {
      updateEveryCount++;
    }
  }

  public File getAdxvImage() {
    return adxvImage;
  }

  protected void setAdxvImage(File newValue) {
    File oldValue = adxvImage;
    adxvImage = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue)) {
      firePropertyChange(Adxvctl.PROPNAME_ADXV_IMAGE, oldValue, newValue);
      writeToAdxvControlFile(newValue);
    }
  }

  public void updateNow() {
    File nowFile = getCollectedImage();
    if (nowFile == null) return;
    enqueueAdxvImageUpdate(nowFile, 0);
    updateEveryCount = 1;
  }

  public boolean isUpdateEveryEnabled() {
    return updateEveryEnabled;
  }

  public void setUpdateEveryEnabled(boolean newValue) {
    boolean oldValue = updateEveryEnabled;
    updateEveryEnabled = newValue;
    if (oldValue != newValue) {
      firePropertyChange(Adxvctl.PROPNAME_UPDATE_EVERY_ENABLED, oldValue,
        newValue);
      if (!newValue) writtenToDiskMonitor.clearQueue();
      if (newValue) updateEveryCount = updateEveryInterval;
    }
  }

  public int getUpdateEveryInterval() {
    return updateEveryInterval;
  }

  public void setUpdateEveryInterval(int newValue) {
    if (newValue < 1) return;
    int oldValue = updateEveryInterval;
    updateEveryInterval = newValue;
    if (oldValue != newValue) {
      firePropertyChange(Adxvctl.PROPNAME_UPDATE_EVERY_INTERVAL, oldValue,
        newValue);
    }
  }

  public String getUserMessage() {
    return userMessage;
  }

  public void setUserMessage(String newValue) {
    if (newValue == null) return;
    String oldValue = userMessage;
    userMessage = newValue;
    if (!oldValue.equals(newValue)) {
      firePropertyChange(Adxvctl.PROPNAME_USER_MESSAGE, oldValue, newValue);
    }
  }

  public void enqueueAdxvImageUpdate(File newFile, long collectTime) {
    boolean enqueuedOk = writtenToDiskMonitor.enqueue(newFile, collectTime);
    if (!enqueuedOk) LOG.error("Adxv image enqueue failed; dropping");
  }

  protected void writeToAdxvControlFile(File aFile) {
    File controlFile = Resources.getAdxvControlFile();
    PrintWriter writer = null;
    try {
      controlFile.createNewFile();
      writer = new PrintWriter(controlFile);
      writer.print((aFile == null) ? 0 : adxvImageCount);
      adxvImageCount++;
      writer.print(' ');
      writer.println((aFile == null)
        ? NONEXISTENT_FILE.getPath()
        : aFile.getPath());
      writer.close();
      setUserMessage("");
    } catch (IOException e) {
      LOG.error("Failed to write to Adxv control file \"" +
        controlFile.getPath() + "\"", e);
      setUserMessage(USER_MESSAGE_PREFIX_ERROR + "The Adxv control file \"" +
        controlFile.getPath() + "\" could not be written.  Try checking " +
        "that the file exists, the file is writable, and that the disk is " +
        "not full.");
    } finally {
      if (writer != null) writer.close();
    }
  }

  protected void collectedImagePathChanged(String newPath) {
    if (newPath.length() == 0) return;
    setCollectedImage(new File(newPath));
  }

  protected void imageWrittenToDisk(File image) {
    setAdxvImage(image);
  }
}
