package imcacat.adxvctl;

import java.awt.Component;
import java.awt.FontMetrics;
import java.io.File;
import java.util.HashMap;
import java.util.Map;

import javax.swing.DefaultListCellRenderer;
import javax.swing.JList;
import javax.swing.SwingUtilities;

/**
 * I render a File object in a list cell.
 */
public class FileCellRenderer extends DefaultListCellRenderer {
  private static final long serialVersionUID = 1;

  public static final int FLAG_NONE          = 0x0000;
  public static final int FLAG_LEFT_TRUNCATE = 0x0001;

  protected boolean leftTruncated;
  protected Map<String, String> truncatedPathCache =
    new HashMap<String, String>();
  protected int listWidth = 0;

  public FileCellRenderer() {
    this(FLAG_NONE);
  }

  public FileCellRenderer(int flags) {
    super();
    leftTruncated = (flags & FLAG_LEFT_TRUNCATE) == FLAG_LEFT_TRUNCATE;
  }

  @Override
  public Component getListCellRendererComponent(JList list, Object value,
      int index, boolean isSelected, boolean cellHasFocus) {
    return super.getListCellRendererComponent(list,
      convertToString((File)value, list), index, isSelected, cellHasFocus);
  }

  public void setListWidth(int listWidth) {
    if (this.listWidth == listWidth) return;
    truncatedPathCache.clear();
    this.listWidth = listWidth;
  }

  protected String convertToString(File value) {
    return (value == null) ? "" : value.getPath();
  }

  protected String convertToString(File value, JList list) {
    return convertToString(value,
      leftTruncated ? list.getFontMetrics(list.getFont()) : null);
  }

  protected String convertToString(File value, FontMetrics metrics) {
    String path = convertToString(value);
    if (!leftTruncated) return path;

    String ellipsis = "\u2026";
    int pathLength = path.length();
    if (pathLength < ellipsis.length()) return path;

    int pathWidth = SwingUtilities.computeStringWidth(metrics, path);
    int arrowWidthApprox = SwingUtilities.computeStringWidth(metrics, "QQQ");
    int listPathWidth = listWidth - arrowWidthApprox;
    if (pathWidth <= listPathWidth) return path;

    if (truncatedPathCache.containsKey(path))
      return truncatedPathCache.get(path);

    int pathEllipsisReplaceWidth = SwingUtilities.computeStringWidth(metrics,
      path.substring(0, ellipsis.length()));
    int ellipsisWidth = SwingUtilities.computeStringWidth(metrics, ellipsis);
    double percentOfPathWidthVisible = (double)listPathWidth /
      (double)(pathWidth - pathEllipsisReplaceWidth + ellipsisWidth);
    int numCharsInPathVisible = (int)((double)pathLength *
      Math.min(1.0, percentOfPathWidthVisible));
    int truncationOffset = Math.max(0, pathLength - numCharsInPathVisible);
    int truncatedPathWidth = SwingUtilities.computeStringWidth(metrics,
      ellipsis + path.substring(truncationOffset));
    int step = (truncatedPathWidth > listPathWidth) ? 1 : -1;
    for (;;) {
      truncatedPathWidth = SwingUtilities.computeStringWidth(metrics,
        ellipsis + path.substring(truncationOffset));
      if (step > 0 && truncatedPathWidth < listPathWidth) break;
      if (step < 0 && truncatedPathWidth > listPathWidth) break;
      if (step < 0 && truncationOffset <= 0) break;
      if (step > 0 && truncationOffset >= pathLength - 1) break;
      truncationOffset += step;
    }
    String truncatedPath = ellipsis + path.substring(truncationOffset);
    truncatedPathCache.put(path, truncatedPath);
    return truncatedPath;
  }
}
