package imcacat.adxvctl;

import java.io.File;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.SwingUtilities;

import org.apache.log4j.Logger;

/**
 * I monitor the file system to determine when a file has been written to
 * disk.
 */
public class WrittenToDiskMonitor implements Runnable {
  private static final Logger LOG = Logger.getLogger(
    WrittenToDiskMonitor.class);

  protected static final long EMPTY_CHECK_DELAY = 0;
  protected static final long MAX_CLOCK_SKEW = 10000;

  protected Object delayObject = new Object();
  protected volatile long checkDelay = EMPTY_CHECK_DELAY;
  protected long nonemptyCheckDelay;
  protected long minUnmodifiedTime;
  protected int maxQueueSize;
  protected long maxTimeInQueue;
  protected boolean notifyOnEdt;
  protected LinkedList<WrittenToDiskQueueItem> queue =
    new LinkedList<WrittenToDiskQueueItem>();
  protected List<WrittenToDiskListener> writtenToDiskListeners =
    new LinkedList<WrittenToDiskListener>();

  public WrittenToDiskMonitor() {
    this(1000, 2000, 1024, 10000, true);
  }

  public WrittenToDiskMonitor(long nonemptyCheckDelay, long minUnmodifiedTime,
      int maxQueueSize, long maxTimeInQueue, boolean notifyOnEdt) {
    this.nonemptyCheckDelay = nonemptyCheckDelay;
    this.minUnmodifiedTime = minUnmodifiedTime;
    this.maxQueueSize = maxQueueSize;
    this.maxTimeInQueue = millisecondsToNanoseconds(maxTimeInQueue);
    this.notifyOnEdt = notifyOnEdt;
  }

  public void run() {
    try {
      for (;;) {
        check();
        synchronized (delayObject) {
          delayObject.wait(checkDelay); /* a spurious wake-up is OK */
        }
      }
    } catch (InterruptedException e) {
      LOG.info("Interrupted; stopping");
    }
  }

  public synchronized boolean enqueue(File aFile) {
    return enqueue(aFile, 0);
  }

  public synchronized boolean enqueue(File aFile, long collectTime) {
    if (aFile == null) throw new IllegalArgumentException("null file");
    if (queue.size() == maxQueueSize) return false;
    if (queueContains(aFile)) return true;
    queue.addLast(new WrittenToDiskQueueItem(aFile, System.nanoTime(),
      collectTime));
    synchronized (delayObject) {
      delayObject.notify();
    }
    return true;
  }

  public synchronized boolean queueContains(File aFile) {
    if (aFile == null) return false;
    for (WrittenToDiskQueueItem each : queue)
      if (aFile.equals(each.getFile())) return true;
    return false;
  }

  public synchronized void clearQueue() {
    queue.clear();
  }

  public synchronized void addWrittenToDiskListener(
      WrittenToDiskListener listener) {
    writtenToDiskListeners.add(listener);
  }

  public synchronized void removeWrittenToDiskListener(
      WrittenToDiskListener listener) {
    writtenToDiskListeners.remove(listener);
  }

  protected synchronized void check() {
    for (Iterator<WrittenToDiskQueueItem> i = queue.iterator();
        i.hasNext();) {
      WrittenToDiskQueueItem each = i.next();
      if (isWrittenToDisk(each)) {
        notifyWrittenToDisk(each.getFile());
        i.remove();
      } else if (hasExceededMaxTimeInQueue(each)) {
        i.remove();
      }
    }
    checkDelay = queue.isEmpty() ? EMPTY_CHECK_DELAY : nonemptyCheckDelay;
  }

  protected synchronized boolean isWrittenToDisk(
      WrittenToDiskQueueItem item) {
    if (!item.getFile().exists()) return false;
    if ((item.getCollectTime() - MAX_CLOCK_SKEW) >
        item.getFile().lastModified())
      return false;
    long unmodifiedTime = System.currentTimeMillis() -
      item.getFile().lastModified();
    return unmodifiedTime >= minUnmodifiedTime;
  }

  protected synchronized boolean hasExceededMaxTimeInQueue(
      WrittenToDiskQueueItem item) {
    long timeInQueue = System.nanoTime() - item.getEnqueueTime();
    return timeInQueue > maxTimeInQueue;
  }

  protected void notifyWrittenToDisk(final File aFile) {
    if (notifyOnEdt && !SwingUtilities.isEventDispatchThread())
      SwingUtilities.invokeLater(new Runnable() {
        public void run() {
          notifyWrittenToDiskFromCurrentThread(aFile);
        }});
    else
      notifyWrittenToDiskFromCurrentThread(aFile);
  }

  protected synchronized void notifyWrittenToDiskFromCurrentThread(
      File aFile) {
    for (WrittenToDiskListener each : writtenToDiskListeners)
      each.writtenToDisk(aFile);
  }

  protected static long millisecondsToNanoseconds(long milliseconds) {
    return milliseconds * 1000000L;
  }
}
